/* $Id: espda_cont_ord.c 2018-01-24 14:15:10 jonaydas $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2018-01-24 14:15:10 +0000 (sat, 24 feb 2017) $
 * $Revision: 202524 $
 * $Name:  $
 */
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>

#include "espda_cont_ord.h"
#include "espda_util_cont.h"
#include "espda_dfs.h"
#include "espda_frame.h"
#include "espda_spec.h"

/**@{*/

/*----------------------------------------------------------------------------*/
/**
 * @defgroup espda_cont_ord
 *
 * Continuum order fit
 */
/*----------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Fit the stellar continuum of a spectral order
  @param    wave     Array of wavelength of the spectral order
  @param    flux     Array of flux of the spectral order
  @param    cont     Array of continuum of the spectral order
  @param    conterr  Array of error on continuum level of unmasked points
  @param    mask     Array of mask values of the spectral order
  @param    lrejt    Low rejection threshold for continuum fitting
  @param    hrejt    High rejection threshold for continuum fitting
  @param    niter    Number of iterations for continuum fitting
  @param    ford     Order of the fitting function
  @return   CPL_ERROR_NONE iff OK

  @doc      
  This function fits the flux of stellar continuum of a given spectral order
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_cont_ord(cpl_array     *wave,
			                  cpl_array     *flux,
 			                  cpl_array     *cont,
 			                  cpl_array  *conterr,
 			                  cpl_array     *mask,
		                      const double  lrejt,
		                      const double  hrejt,
							  const int     niter,
			                  const int      ford)

{
    cpl_errorstate prestate = cpl_errorstate_get();
// Defining the number of elements of the arrays
    cpl_size npoint = cpl_array_get_size (wave);

/*
 * This function fits the continuum flux of a spectral order of a * 
 *  stellar spectrum and gives as output an array with values of  *
 *  the fitted function                                           *
 */

    // We create a constant "cont" array with median values
    // We obtain the median of the array flux
    //printf("###################################\n");
    //printf("############# NEW ORDER SPEC = %i %i %i\n", nord, nspc, ford);
    //printf("###################################\n");

	//printf("############# NUMBER points ORDER = %i \n", (int)npoint);

    cpl_array *flx  = cpl_array_duplicate (flux);
    //cpl_array *fer  = cpl_array_duplicate (fluxerr);

    double cmedf = cpl_array_get_median (flx);
    double cmd = cpl_array_get_median (flx);
    double cme = cpl_array_get_mean (flx);
    if (cmd > 0.) {
    	cmedf = cmd;
    } else {
    	if (cme > 0.) {
    		cmedf = cme;
    	} else {
    		cmedf = (-1.)*cme;
    	}
    }
    //printf("MEDIAN FLUX points: %f \n", cmedf);
    //const double cmdf = cpl_array_get_median (flx);
    //printf("MEAN FLUX points: %f \n", cmdf);

    //double cmax = cpl_array_get_max (flx);
    //printf("Maximum FLUX points: %f \n", cmax);
    //double cmin = cpl_array_get_min (flx);
    //printf("Minimum FLUX points: %f \n", cmin);
    //double cmedi = cpl_array_get_median (flx);
    //printf("MEDIAN FLUX points: %f \n", cmedi);
    //double cmean = cpl_array_get_mean (flx);
    //printf("Mean FLUX points: %f \n", cmean);
    //double cstd = cpl_array_get_stdev (flx);
    //printf("STDDEV FLUX points: %f \n", cstd);
    //double cfer = cpl_array_get_mean (fer);
    //printf("MEAN ERROR FLUX points: %f \n", cfer);
    //printf("\n");

    // We initialize the fitting process
    //printf("Order Fitting Function: %i \n", forder);

    // We initiate the arrays

    cpl_array *wax  = cpl_array_duplicate (wave);
    cpl_array *nflx  = cpl_array_duplicate (flx);
    // Arrays with only unmasked points evaluated
    cpl_array *ctx_i  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);
    cpl_array *ctx_e  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);
    cpl_array *ctx_in  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);
    cpl_array *ctx_err  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);
    // Arrays with all points evaluated: initial Continuum fit
    cpl_array *ctx_o  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);
    cpl_array *ctx_out  = cpl_array_new (npoint, CPL_TYPE_DOUBLE);

    // We create a new "normalized" array by dividing the array by its MEDIAN
    
    cpl_array_divide_scalar (nflx, cmedf);
    //cpl_array_divide_scalar (nflx, cmdf);
    //cmax = cpl_array_get_max (nflx);
    //printf("Maximum NFLUX0 points: %f \n", cmax);
    //cmin = cpl_array_get_min (nflx);
    //printf("Minimum NFLUX0 points: %f \n", cmin);
    //cmedi = cpl_array_get_median (nflx);
    //printf("MEDIAN NFLUX0 points: %f \n", cmedi);
    //cmean = cpl_array_get_mean (nflx);
    //printf("Mean NFLUX0 points: %f \n", cmean);
    //cstd = cpl_array_get_stdev (nflx);
    //printf("STDDEV NFLUX0 points: %f \n", cstd);
    //printf("\n");

    //We initiate variables
    // We define the new mask
    //  - mask = 1 for selected points
    //  - mask = 0 for rejected points
    // We initially mask points with FLUX / MEDIAN(FLUX) < 0.

    int nrej = 0 ;
    cpl_array *msx  = cpl_array_new (npoint, CPL_TYPE_INT);
    for (int j = 0; j < (int)npoint; j++) {
      int cmsx = 1;
      cpl_array_set_int (msx, j, cmsx);
      double cnflx = cpl_array_get_double (nflx, j, NULL);
      if (cnflx < 0.) {
        int cmsxa = 0; 
        cpl_array_set_int (msx, j, cmsxa);
        nrej++;
      }    
    }

    int ni = (int)npoint-nrej ;
    //printf("Initial Number of points: %i \n", (int)npoint);
    //printf("Initial Number of selected points: %i \n", ni);
    //printf("Initial Number of discarded points: %i \n", nrej);

    // We first normalized the spectra using a first order polynomial
    //printf("First, we do 1st order \n");
    //printf("\n");

    double cwax = cpl_array_get_double (wax, 0, NULL);
    int fordi = 0;
    if (cwax > 540.){
       fordi = 1;
    } else {
       fordi = 0;
    }
    /*Telluric oxygen features */
    if (cwax > 759. && cwax < 762.){
       fordi = 0;
    }

    const int ncoefi = (int)fordi +1 ;
    cpl_array *xcf_i  = cpl_array_new (ncoefi, CPL_TYPE_DOUBLE);

    //printf("############# CMED %f FORDER in %i %i\n", cmedf, fordi, ford);

    cpl_ensure_code(espda_cont_ord_fit(wax, nflx, msx,
			   fordi, ni, xcf_i, ctx_i, ctx_e)
               == CPL_ERROR_NONE, cpl_error_get_code());

    // We select only those flux points, ford, that remain in the range
    // [fcont-fstd*lrej,fcont+fstd*hrej] where fcont is the flux of the fitting
    // function, fstd is the STDDEV of the normalized order spectrum

    //cpl_ensure_code(espda_cont_mask(wax, nflx, ctx_i,
    //		msx, hrejt, lrejt)
    //		== CPL_ERROR_NONE, cpl_error_get_code());

    //Deriving the number of mask points:

    int nrj= 0;

       for (int j = 0; j < (int)npoint; j++) {
       int cmsx = cpl_array_get_int (msx, j, NULL);
       if ((int)cmsx == 0) {
         nrj++;
       }
     }
    nrej = nrj;

    // We evaluate the fitting function
    //printf("############# FORDER out %i %i\n", fordi, ford);
    //for (int j = 0; j < (int)fordi+1; j++) {
    // double ccoef = cpl_array_get_double (xcf_i, j, NULL);
    // printf("Coeff0 %i %f\n", j, ccoef);
    //}
    cpl_ensure_code(espda_cont_ord_eval(wax, xcf_i,
                   ctx_o) == CPL_ERROR_NONE, cpl_error_get_code());

    // We save the CONTINUUM and Nflux arrays of the first order continuum fitting

     cpl_array_divide (nflx, ctx_o);
     //cmax = cpl_array_get_max (nflx);
     //printf("Maximum NFLUX1 points: %f \n", cmax);
     //cmin = cpl_array_get_min (nflx);
     //printf("Minimum NFLUX1 points: %f \n", cmin);
     //cmedi = cpl_array_get_median (nflx);
     //printf("MEDIAN NFLUX1 points: %f \n", cmedi);
     //cmean = cpl_array_get_mean (nflx);
     //printf("Mean NFLUX1 points: %f \n", cmean);
     //cstd = cpl_array_get_stdev (nflx);
     //printf("STDDEV NFLUX1 points: %f \n", cstd);
     //printf("\n");

     // Defining the number of elements of the arrays

     ni = (int)npoint-nrej ;
     int nf = 0 ;
     //printf("1stOrder Number of points: %i \n", (int)npoint);
     //printf("1stOrder Number of selected points: %i \n", ni);
     //printf("1stOrder Number of discarded points: %i \n", nrej);

    // We start the loop over Niter and Npoint

    int i = 0 ;
    int ncoef = (int)ford+1 ;
    cpl_array *xcf_in  = cpl_array_new (ncoef, CPL_TYPE_DOUBLE);

    while (i < (int)niter+1 && (int)nf < (int)ni)  {

    	ni = (int)npoint-nrej;

    // We perform the fit using polynomial of order "forder"

      //printf("Iteration Number: %i \n", i);
      //printf("Niter: %i Order: %i Orderb: %i \n", i, ford, fordi);

      cpl_ensure_code(espda_cont_ord_fit(wax, nflx, msx,
			   ford, ni, xcf_in, ctx_in, ctx_err)
               == CPL_ERROR_NONE, cpl_error_get_code());

    // We select only those flux points, ford, that remain in the range
    // [fcont-fstd*lrej,fcont+fstd*hrej] where fcont is the flux of the fitting
    // function, fstd is the STDDEV of the normalized order spectrum

      if (i < (int)niter) {
    	cpl_ensure_code(espda_cont_mask(wax, nflx,
    			ctx_in, msx, hrejt, lrejt)
    			== CPL_ERROR_NONE, cpl_error_get_code());
        //Deriving the number of mask points:
        nrj= 0;
        for (int j = 0; j < (int)npoint; j++) {
           int cmsx = cpl_array_get_int (msx, j, NULL);
           if ((int)cmsx == 0) {
             nrj++;
           }
         }
        nrej = nrj;
      }
      //printf("Number of rejected points: %i \n", nrej);

    //Closing loop over Npoint and Niter
      i++;
      nf = (int)npoint-nrej ;
    }

    //Creating the continuun for all points

     // We evaluate the fitting function
     //for (int j = 0; j < (int)ford+1; j++) {
     // double ccoef = cpl_array_get_double (xcf_in, j, NULL);
     // printf("Coeff1 %i %f\n", j, ccoef);
     //}

     cpl_ensure_code(espda_cont_ord_eval(wax, xcf_in,  
                    ctx_out) == CPL_ERROR_NONE, cpl_error_get_code());

     //Saving final arrays of CONTINUUM and MASK

     for (int j = 0; j < (int)npoint; j++) {
       int cmsx = cpl_array_get_int (msx, j, NULL);
       cpl_array_set_int (mask, j, cmsx);
       double cctx = cpl_array_get_double (ctx_out, j, NULL);
       double cctxi = cpl_array_get_double (ctx_o, j, NULL);
       double ccont = cctx * cctxi * cmedf;
       cpl_array_set_double (cont, j, ccont);
       if ((int)cmsx == 1) {
         double cctxe = cpl_array_get_double (ctx_err, j, NULL);
         double ccerr = cctxe * cctxi * cmedf;
       cpl_array_set_double (conterr, j, ccerr);
       }
     }

    // We delete all the arrays to release and clean the memory

    cpl_array_delete (msx);
    cpl_array_delete (ctx_in);
    cpl_array_delete (ctx_err);
    cpl_array_delete (ctx_out);
    cpl_array_delete (ctx_o);
    cpl_array_delete (ctx_i);
    cpl_array_delete (ctx_e);
    cpl_array_delete (wax);
    cpl_array_delete (nflx);
    cpl_array_delete (flx);
    cpl_array_delete (xcf_in);
    cpl_array_delete (xcf_i);

    if (!cpl_errorstate_is_equal(prestate)) {
      return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to perform the continuum fitting");
    }

    return CPL_ERROR_NONE;
}

/**@}*/
