
/* 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Id: espda_param.c 161538 2014-05-22 13:56:10Z gcupani $
 * $Author: gcupani $
 * $Date: 2014-05-22 15:56:10 +0200 (Thu, 22 May 2014) $
 * $Revision: 161538 $
 * $Name: not supported by cvs2svn $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_param.h"

/*----------------------------------------------------------------------------*/
/**
  @defgroup  espda_param  Parameter I/O
 
  Utilities to handle parameters.
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  DEPRECATED!
do not delete yet, it's used by macros!
Change espda_param_TYPE to espda_param_load
 */
/*----------------------------------------------------------------------------*/

double espda_param_adjust(cpl_type type, double value_in, double min, 
    double max) {
     double value_out = value_in;
     if (value_in < min) {
         value_out = min;
         if (type == CPL_TYPE_DOUBLE) {
             cpl_msg_warning(cpl_func, " Value %f is too low. Redefined to %f.",
                             value_in, value_out);
         }
         if (type == CPL_TYPE_INT) {
             cpl_msg_warning(cpl_func, " Value %i is too low. Redefined to %i.",
                             (int)value_in, (int)value_out);
         }
     }    
     if (value_in > max) {
         value_out = max;
         if (type == CPL_TYPE_DOUBLE) {
             cpl_msg_warning(cpl_func, " Value %f is too high. Redefined to "
                             "%f.", value_in, value_out);
         }
         if (type == CPL_TYPE_INT) {
             cpl_msg_warning(cpl_func, " Value %i is too high. Redefined to "
                             "%i.", (int)value_in, (int)value_out);
         }
     }   
     return value_out; 
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Extract a string parameter into an array.
  @param   param  A string parameter.
  @param   sep    The separator to parse string entries.
  @param   start  The starting string entry.
  @param   step   The step to parse string entries.
  @param   type   The type of the array.
  @return  An array with the string entries.
  
  This function parses a string parameter @a param consisting in a list of 
  entries identified by the separator @a sep. The parsing starts from entry 
  @a start and uses a step @a step. An array of type @a type is produced from
  the extracted entries.
  
  If @a param or @a sep are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_array *espda_param_extract(const char *param, char *sep, 
    cpl_size start, cpl_size step, cpl_type type)  {
    espda_util_error(param == NULL, CPL_ERROR_NULL_INPUT, NULL);
    espda_util_error(sep   == NULL, CPL_ERROR_NULL_INPUT, NULL);
    cpl_array *temp = cpl_array_new(LEN_STRING, type);
    cpl_array *out = NULL;
    //char *tok;
    char str[LEN_STRING];
    strcpy(str, param);              
    if (strcmp(str, "")) {
      char *tok = (char *)malloc(sizeof(char) * LEN_STRING);
        if (strstr(str, sep)) {
	    tok = strtok(str, sep);
        }
        char *null;
        int i = 0;
        while (tok != NULL) {    
            if (type == CPL_TYPE_INT) {
                cpl_array_set_int   (temp, i, (int)strtod(tok, &null));
            }
            if (type == CPL_TYPE_DOUBLE) {
                cpl_array_set_double(temp, i, strtod(tok, &null));
            }
            if (type == CPL_TYPE_STRING) {
                cpl_array_set_string(temp, i, tok);
            }
            tok = strtok(NULL, sep);    
            i++;
        }   
        if (i > 0) {
            out = cpl_array_new(i, type);
            int k = 0;
            for (int j = start; j < i; j = j + step) {
                if (type == CPL_TYPE_INT) {
                    cpl_array_set_int(out, k, cpl_array_get_int(temp, j, NULL));
                }
                if (type == CPL_TYPE_DOUBLE) {
                    cpl_array_set_double(out, k, cpl_array_get_double(temp, j, 
                        NULL));
                }
                if (type == CPL_TYPE_STRING) {
                    cpl_array_set_string(out, k, cpl_array_get_string(temp, j));
                }
                k++;
            }
            cpl_array_set_size(out, k);
        }
	cpl_free(tok);
    }
    cpl_array_delete(temp);
    return out;
}

/*----------------------------------------------------------------------------*/
/**
  @brief   Load a parameter from a parameter list.
  @param   list     A parameter list.
  @param   name     The parameter's name.
  @param   ...      A pointer to the parameter value and an optional range.
  @return  CPL error code.
  
  This function scans a parameter list @a list and finds the parameter @a name.
  A pointer to the parameter value is issued and may be dereferenced to an 
  integer, a double, or a string variable. If the parameter is an integer or a
  double, the allowed range must be specified after the pointer as a pair of
  values (minimum and maximum respectively). If the parameter value falls 
  outside this range, it is changed to the nearest extreme of the range itself.
  
  If @a list or @a name are NULL, an error is issued. If an integer or double 
  parameter has been changed to be within the allowed range and the verbosity
  level is greater than 0, a warning is issued. 
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_param_load(const cpl_parameterlist *list, const char *name,
    ...) {
    extern int v_level;
    espda_util_error(list == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(name == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    const cpl_parameter *param = cpl_parameterlist_find_const(list, name);
    const char *alias = cpl_parameter_get_alias(param, CPL_PARAMETER_MODE_CLI);
    cpl_type type = cpl_parameter_get_type(param);
    va_list ap;
    va_start(ap, name);
    if (type == CPL_TYPE_BOOL) {
        int *value_bool = va_arg(ap, int *);
        *value_bool = cpl_parameter_get_bool(param);
    }
    else if (type == CPL_TYPE_DOUBLE) {
        double *value_double = va_arg(ap, double *);
        double  min_double   = va_arg(ap, double);
        double  max_double   = va_arg(ap, double);
        *value_double = cpl_parameter_get_double(param);
        espda_param_adjust2(*value_double, min_double, max_double, alias, %f);
    }
    else if (type == CPL_TYPE_INT) {
        int *value_int = va_arg(ap, int *);
        int  min_int   = va_arg(ap, int);
        int  max_int   = va_arg(ap, int);
        *value_int = cpl_parameter_get_int(param);
        espda_param_adjust2(*value_int, min_int, max_int, alias, %i);
    }
    else if (type == CPL_TYPE_STRING) {
        char **value_string = va_arg(ap, char **);
        const char *temp = cpl_parameter_get_string(param);
        strcpy(*value_string, temp);
    }
    va_end(ap);
    return cpl_error_get_code();
} 

/*----------------------------------------------------------------------------*/
/**
  @brief   Replace a pattern in a string parameter.
  @param   param  A string parameter.
  @param   find   The pattern to find.
  @param   repl   The replacement.
  @param   out    The string parameter with the pattern replaced.
  
  This function finds a pattern @a find in a parameter @a param, and replaces it
  with a replacement @a repl. A parameter @a out with the pattern replaced is
  produced.
  
  If @a param, @a find, or @a repl are NULL, an error is issued.
 */
/*----------------------------------------------------------------------------*/

cpl_error_code espda_param_replace(const char *param, char *find,
				   char *repl, char *out){//[LEN_STRING]) {
    espda_util_error(param == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(find  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    espda_util_error(repl  == NULL, CPL_ERROR_NULL_INPUT, cpl_error_get_code());
    char  replace_str[LEN_STRING];
    strcpy(replace_str, param);
    strcpy(out, "");
    char *replace_tok = strtok(replace_str, find);
    while (replace_tok != NULL) {
        if (strcmp(out, "")) {
            strcat(out, repl);
        }
        strcat(out, replace_tok);
        replace_tok = strtok(NULL, find);
    }
    return cpl_error_get_code();
}

/**@}*/
