/* $Id: espda_util_cont.c 2018-05-02 14:15:10 jonaydas $
 * 
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: jonaydas $
 * $Date: 2018-05-02 14:15:10 +0000 (sat, 02 may 2018) $
 * $Revision: 202524 $
 * $Name:  $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

/*-----------------------------------------------------------------------------
                                   Includes
 -----------------------------------------------------------------------------*/

#include "espda_util_cont.h"
#include <gsl/gsl_version.h>

/*----------------------------------------------------------------------------*/
/**
  @defgroup   espda_util_cont  Utilities for continuum fitting
 */
/*----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Perform the continuum fitting function
  @param    wave_in    Wavelength of the spectral order
  @param    flux_in    Flux of the spectral order
  @param    mask_in    Mask values of the spectral order
  @param    forder     Order of fitting polynomial function
  @param    nfin       Number of unmasked fitting points
  @param    coef_out   Polynomial coefficients
  @param    cont_out   Continuum of the spectral order
  @param    cont_err   Error on continuum level of unmasked points
  @return   CPL_ERROR_NONE  iff OK
        
  This function fits a polynomial with a given number of coefficients, forder+1, 
typically less than 5, in a wavelength array, wave, and the flux of the continuum 
using the unmasked flux points

  We use the Robust linear regression:
Ordinary least squares (OLS) models are often heavily influenced by the presence 
of outliers. Outliers are data points which do not follow the general trend of the 
other observations, although there is strictly no precise definition of an outlier.
Robust linear regression refers to regression algorithms which are robust to outliers.
The most common type of robust regression is M-estimation.

Function: 
gsl_multifit_robust_workspace * gsl_multifit_robust_alloc (const gsl_multifit_robust_type * T, const size_t n, const size_t p)

This function allocates a workspace for fitting a model to n observations using p 
parameters. The type T specifies the function \psi and can be selected from the 
following choices.

Robust type: gsl_multifit_robust_bisquare
This is Tukey's biweight (bisquare) function and is a good general purpose choice for 
robust regression. The weight function is given by w(e) = (1 - e^2)^2
and the default tuning constant is t = 4.685.

Function: void gsl_multifit_robust_free (gsl_multifit_robust_workspace * w)
This function frees the memory associated with the workspace w.

Function: int gsl_multifit_robust (const gsl_matrix * X, const gsl_vector * y, gsl_vector * c, gsl_matrix * cov, gsl_multifit_robust_workspace * w)
This function computes the best-fit parameters c of the model y = X c for the observations 
y and the matrix of predictor variables X, attemping to reduce the influence of outliers 
using the algorithm outlined above. The p-by-p variance-covariance matrix of the model 
parameters cov is estimated as \sigma^2 (X^T X)^{-1}, where \sigma is an approximation of 
the residual standard deviation using the theory of robust regression. Special care must 
be taken when estimating \sigma and other statistics such as R^2, and so these are computed internally and are available by calling the function gsl_multifit_robust_statistics.

Function: gsl_multifit_robust_stats gsl_multifit_robust_statistics (const gsl_multifit_robust_workspace * w)
This function returns a structure containing relevant statistics from a robust regression. 
The function gsl_multifit_robust must be called first to perform the regression and 
calculate these statistics. The returned gsl_multifit_robust_stats structure contains the 
following fields.

- double sigma_ols This contains the standard deviation of the residuals as computed from 
ordinary least squares (OLS).
- double sigma_mad This contains an estimate of the standard deviation of the final 
residuals using the Median-Absolute-Deviation statistic
- double sigma_rob This contains an estimate of the standard deviation of the final 
residuals from the theory of robust regression (see Street et al, 1988).
- double sigma This contains an estimate of the standard deviation of the final residuals 
by attemping to reconcile sigma_rob and sigma_ols in a reasonable way.
- double Rsq This contains the R^2 coefficient of determination statistic using the estimate
sigma.
- double adj_Rsq This contains the adjusted R^2 coefficient of determination statistic 
using the estimate sigma.
- double rmse This contains the root mean squared error of the final residuals
- double sse This contains the residual sum of squares taking into account the robust 
covariance matrix.
- size_t dof This contains the number of degrees of freedom n - p
- size_t numit Upon successful convergence, this contains the number of iterations performed
- gsl_vector * weights This contains the final weight vector of length n
- gsl_vector * r This contains the final residual vector of length n, r = y - X c

 */

/*----------------------------------------------------------------------------*/

cpl_error_code espda_cont_ord_fit(cpl_array  *wave_in,
 			                      cpl_array  *flux_in,
 			                      cpl_array  *mask_in,
		                          const int      ford,
		                          const int      nfin,
 			                      cpl_array *coef_out,
                                  cpl_array *cont_out,
                                  cpl_array *cont_err)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    const int ncoef = ford+1;
    //printf("Order Fit: %i \n", ford);
    cpl_size npoint = cpl_array_get_size (wave_in);
    cpl_array *wax_in = cpl_array_new(nfin, CPL_TYPE_DOUBLE);
    cpl_array *flx_in = cpl_array_new(nfin, CPL_TYPE_DOUBLE);
    cpl_array *ctx_in = cpl_array_new(nfin, CPL_TYPE_DOUBLE);
    cpl_array *ctx_err = cpl_array_new(nfin, CPL_TYPE_DOUBLE);

    // We define the fitting wave and flux using the mask points
    // The arrays wave_in, flux_in and cont_in contain all the 
    // data points, even those with MASK = 0, we then select only
    // those with MASK = 1, and therefore the size of wax_in, flx_in
    // and ctx_in should be after first iteration smaller than npoint

    int k=0;
    for (int j = 0; j < (int)npoint; j++) {
      int cmsx = cpl_array_get_int (mask_in, j, NULL);
      if ((int)cmsx == 1) {
        double cwax = cpl_array_get_double (wave_in, j, NULL);
        cpl_array_set_double (wax_in, k, cwax);
        double cflx = cpl_array_get_double (flux_in, j, NULL);
        cpl_array_set_double (flx_in, k, cflx);
        k++;
      }         
    }

    // We start the gsl fitting workspace
       
    //    printf("Initiating GSL Robust fitting \n");
    gsl_multifit_robust_workspace *wk_sp =
    //gsl_multifit_robust_alloc (gsl_multifit_robust_bisquare, nfin, ncoef);
    gsl_multifit_robust_alloc (gsl_multifit_robust_ols, nfin, ncoef);
    gsl_vector *yvec = gsl_vector_alloc(nfin);
    gsl_vector *xcoef = gsl_vector_alloc(ncoef);
    gsl_matrix *xmtx = gsl_matrix_alloc(nfin, ncoef);
    gsl_matrix *cmtx = gsl_matrix_alloc(ncoef, ncoef);

    //We create the GSL X matrix of the fitting function y=Xc where y and c are
    //GSL vectors: the GSL X matrix is defined for instance as follows:
    //The next program performs a quadratic fit y = c_0 + c_1 x + c_2 x^2 to a 
    //weighted dataset using the generalised linear fitting function 
    //gsl_multifit_wlinear. The model matrix X for a quadratic fit is given by,

    //X = [ 1   , x_0  , x_0^2 ;
    //      1   , x_1  , x_1^2 ;
    //      1   , x_2  , x_2^2 ;
    //      ... , ...  , ...   ]
    //where the column of ones corresponds to the constant term c_0. The two 
    //remaining columns corresponds to the terms c_1 x and c_2 x^2.

    //    printf("Building GSL input fitting vector and matrixes \n");
    for (int k = 0; k < (int)ncoef; k++) {
      cpl_array *wax_out = cpl_array_duplicate (wax_in);
      double cexp = (double)k;
      cpl_array_power (wax_out, cexp);
      for (int j = 0; j < (int)nfin; j++) {
         double cwax = cpl_array_get_double(wax_out, j, NULL);
         double cflx = cpl_array_get_double(flx_in, j, NULL);
         gsl_vector_set(yvec, j, cflx);
         gsl_matrix_set(xmtx, j, k, cwax);
      }
      cpl_array_delete(wax_out);
    }

    // We initialize the fit

//Function: int gsl_multifit_robust (const gsl_matrix * X, const gsl_vector * y, 
//gsl_vector * c, gsl_matrix * cov, gsl_multifit_robust_workspace * w)
//This function computes the best-fit parameters c of the model y = X c for the
//observations y and the matrix of predictor variables X, attemping to reduce the 
//influence of outliers using the algorithm outlined above. The p-by-p variance-covariance
//matrix of the model parameters cov is estimated as \sigma^2 (X^T X)^{-1}, where \sigma 
//is an approximation of the residual standard deviation using the theory of robust 
//regression. Special care must be taken when estimating \sigma and other statistics 
//such as R^2, and so these are computed internally and are available by calling the 
//function gsl_multifit_robust_statistics.

//    printf("Executing GSL Robust fitting \n");
    gsl_multifit_robust (xmtx, yvec, xcoef, cmtx, wk_sp);

//Function: int gsl_multifit_robust_est (const gsl_vector * x, const gsl_vector * c, 
// const gsl_matrix * cov, double * y, double * y_err)
//This function uses the best-fit robust regression coefficients c and their covariance
//matrix cov to compute the fitted function value y and its standard deviation y_err for 
//the model y = x.c at the point x.

     // We evaluate the fitting function
    
    gsl_vector *xvec = gsl_vector_alloc(ncoef);

    //    printf("Creating Continuum FIRST fitting function \n");
    for (int j = 0; j < (int)nfin; j++) {
      for (int k = 0; k < (int)ncoef; k++) {
	    double cwax = gsl_matrix_get (xmtx, j, k);
        gsl_vector_set (xvec, k, cwax);
      }
      double  cvec ;
      double  cvec_err ;
      gsl_multifit_robust_est (xvec, xcoef, cmtx, &cvec, &cvec_err);

    // We write the result in the output

      cpl_array_set_double (ctx_in, j, cvec);
      cpl_array_set_double (ctx_err, j, cvec_err);
    }

    for (int k = 0; k < (int)ncoef; k++) {
      double ccoef = gsl_vector_get (xcoef, k);
      cpl_array_set_double (coef_out, k, ccoef);
      //      printf("Coefficient: %f \n", ccoef);
    }
    
    //    printf("Creating Continuum SECOND fitting function \n");
    int h = 0;
    int l = 0;
    for (int j = 0; j < (int)npoint; j++) {
      int cmsx = cpl_array_get_int (mask_in, j, NULL);
      if ((int)cmsx == 1) {
        double cflxa = cpl_array_get_double (ctx_in, h, NULL);
        cpl_array_set_double (cont_out, j, cflxa);
        double cerra = cpl_array_get_double (ctx_err, h, NULL);
        cpl_array_set_double (cont_err, j, cerra);
        h++;
      } 
      else {
	double cflxb = 0.;
        cpl_array_set_double (cont_out, j, cflxb);
	double cerrb = 0.;
        cpl_array_set_double (cont_err, j, cerrb);
        l++;
      }
    }
    //    printf("Number of fitting points h: %i \n", h);
    //    printf("Number of rejected points l: %i \n", l);
    //    printf("Number of fitting points nfin: %i \n", nfin);
    //    int nnew = h + l ;
    //    printf("Number of points nnew: %i \n", nnew);
     
 // We finilize the workspace of Robust polynomial fit
    
    //     printf("Release GSL memory \n");

    gsl_multifit_robust_free (wk_sp);
    gsl_vector_free (yvec);
    gsl_matrix_free (xmtx);
    gsl_matrix_free (cmtx);
    gsl_vector_free (xcoef);
    gsl_vector_free (xvec);

// We delete all the arrays to release and clean the memory

//    printf("Release CPL memory \n");

    cpl_array_delete (wax_in);
    cpl_array_delete (flx_in);
    cpl_array_delete (ctx_in);
    cpl_array_delete (ctx_err);

     // We finalize the function
    //     printf("Finalize the function \n");
     if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to perform the fit of the continuum flux...");
     }

     return CPL_ERROR_NONE;
}


/*----------------------------------------------------------------------------*/
/**
  @brief    Evaluate the continuum fitting function
  @param    wax     Array of wavelength of the spectral order
  @param    cfx     Array of polynomial coefficients
  @param    ctx     Array of continuum of the spectral order
  @return   CPL_ERROR_NONE  iff OK
        
  This function evaluates a polynomial with a given number of coefficients, forder+1, 
typically less than 5, in a wavelength array, wave, and gives the continuum fitting 
function array, cont as output.
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_cont_ord_eval(cpl_array     *wax,
 			                       cpl_array     *cfx,
                                   cpl_array     *ctx)

{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (wax);
    cpl_size   ncoef = cpl_array_get_size (cfx);
    cpl_array *ctx_out = cpl_array_new (npoint, CPL_TYPE_DOUBLE);

    // Loop on number of coefficients
    for (int j = 0; j < (int)ncoef; j++) {
      double cexp = (double)j;
      double ccoef = cpl_array_get_double (cfx, j, NULL);
      //      printf("Coefficient: %f \n", ccoef);
      cpl_array *ctx_in = cpl_array_duplicate (wax);
      cpl_array_power (ctx_in, cexp);
      cpl_array_multiply_scalar (ctx_in, ccoef);
      //      cpl_array_add (ctx_out, ctx_in);
      for (int k = 0; k < (int)npoint; k++) {
        double ccxta = cpl_array_get_double (ctx_out, k, NULL);
        double ccxtb = cpl_array_get_double (ctx_in, k, NULL);
        double ccxtc = ccxta + ccxtb ;
	  //      printf("Continuum Value pixel: %i %f \n", k, ccxta);
	  //      printf("Continuum Value pixel: %i %f \n", k, ccxtb);
	  //      printf("Continuum Value pixel: %i %f \n", k, ccxtc);
        cpl_array_set_double(ctx_out, k, ccxtc);  
      }
      //      double ccxta = cpl_array_get_double (ctx_out, 1000, NULL);
      //      printf("Continuum Value pixel 1000: %f \n", ccxta);
      //      double ccxtb = cpl_array_get_double (ctx_in, 1000, NULL);
      //      printf("Continuum Value pixel 1000: %f \n", ccxtb);
      cpl_array_delete(ctx_in);
    }

    // We copy the result to the output cont array
    // Save the output cont array into the output array
    for (int j = 0; j < (int)npoint; j++) {
     double cctx = cpl_array_get_double(ctx_out, j, NULL);
     cpl_array_set_double(ctx, j, cctx);  
    }

    // We release the memory by deleting the remaining arrays
    cpl_array_delete(ctx_out);

    if (!cpl_errorstate_is_equal(prestate)) {
        return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
               "Unable to evaluate the continuum fitting function...");
    }

    return CPL_ERROR_NONE;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Masking points for continuum fitting function
  @param    wax     Array of wavelength of the spectral order
  @param    flx     Array of flux of the spectral order
  @param    ctx     Array of continuum of the spectral order
  @param    msx     Array of mask values of the spectral order
  @param    hrejt   High rejection threshold for continuum fitting
  @param    lrejt   Low rejection threshold for continuum fitting
  @return   CPL_ERROR_NONE  iff OK

  This function creates a mask equal to 0 if the wavelength/flux
  point is rejected.
  The condition for rejection is :
     flx[i]/ctx[i] > 1. + cstd * hrejt
     flx[i]/ctx[i] < 1. - cstd * lrejt
  IF the flux point is not rejected then mask is equal to 1
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_cont_mask(cpl_array     *wax,
		                       cpl_array     *flx,
		                       cpl_array     *ctx,
							   cpl_array     *msx,
							   const double hrejt,
		                       const double lrejt)
{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size   npoint = cpl_array_get_size (flx);

    /*Telluric oxygen features */

    for (int j = 0; j < (int)npoint; j++) {
    	double cwax = cpl_array_get_double (wax, j, NULL);
        if (cwax > 759.3 && cwax < 761.7){
        	int cmsx = 0;
        	cpl_array_set_int(msx, j, cmsx);
    	}
        if (cwax > 761.9 && cwax < 768.0){
        	int cmsx = 0;
        	cpl_array_set_int(msx, j, cmsx);
    	}
    }

    //Deriving the number of mask points:
    int nrj= 0;
    for (int j = 0; j < (int)npoint; j++) {
       int cmsx = cpl_array_get_int (msx, j, NULL);
       if ((int)cmsx == 0) {
         nrj++;
       }
     }
    int nrej = nrj;
    int ni = (int)npoint-nrej ;

    // We create the normalized flux of only selected unmasked points

     //printf("Computing Normalized flux with MASK = 1 \n");
     //printf("Number of DISCARDED points BEFORE: %i \n", nrej);
     cpl_array *nflx  = cpl_array_new (ni, CPL_TYPE_DOUBLE);
     cpl_array *mflx  = cpl_array_new (ni, CPL_TYPE_DOUBLE);
     cpl_array *fctx  = cpl_array_new (ni, CPL_TYPE_DOUBLE);

     int k = 0 ;
     for (int j = 0; j < (int)npoint; j++) {
       int cmsx = cpl_array_get_int (msx, j, NULL);
       if ((int)cmsx == 1) {
         double cctx = cpl_array_get_double (ctx, j, NULL);
         double cflx = cpl_array_get_double (flx, j, NULL);
         double cnfx = cflx / cctx;
         cpl_array_set_double (nflx, k, cnfx);
         cpl_array_set_double (mflx, k, cflx);
         cpl_array_set_double (fctx, k, cctx);
         k++;
       }
     }

    // We compute the STDDEV of the normalized order spectrum

        //printf("Values for NON-DISCARDING points\n");
        //double cmax = cpl_array_get_max (nflx);
        //printf("Maximum NFLUX points: %f \n", cmax);
        //double cmin = cpl_array_get_min (nflx);
        //printf("Minimum NFLUX points: %f \n", cmin);
        //double cmedi = cpl_array_get_median (nflx);
        //printf("MEDIAN NFLUX points: %f \n", cmedi);
        //double cmean = cpl_array_get_mean (nflx);
        //printf("Mean NFLUX points: %f \n", cmean);
        double cstd = cpl_array_get_stdev (nflx);
        //printf("STDDEV NFLUX points: %f \n", cstd);
        //printf("\n");

        //cmax = cpl_array_get_max (mflx);
        //printf("Maximum FLUX points: %f \n", cmax);
        //cmin = cpl_array_get_min (mflx);
        //printf("Minimum FLUX points : %f \n", cmin);
        //cmedi = cpl_array_get_median (mflx);
        //printf("MEDIAN FLUX points : %f \n", cmedi);
        //printf("\n");

        //cmax = cpl_array_get_max (fctx);
        //printf("Maximum CONT points: %f \n", cmax);
        //cmin = cpl_array_get_min (fctx);
        //printf("Minimum CONT points: %f \n", cmin);
        //cmedi = cpl_array_get_median (fctx);
        //printf("MEDIAN CONT points: %f \n", cmedi);
        //printf("\n");

        //We delete the normalized flux of non-discarded points
        cpl_array_delete (nflx);
        cpl_array_delete (mflx);
        cpl_array_delete (fctx);

    // Masking flux points
    for (int j = 0; j < (int) npoint; j++) {
      int cmsxb = cpl_array_get_int(msx, j, NULL);
      if ((int) cmsxb == 1) {
        double cctx = cpl_array_get_double(ctx, j, NULL);
        double cflx = cpl_array_get_double(flx, j, NULL);
        double cnfx = cflx / cctx;
        if (cnfx > 1. + cstd * hrejt) {
          int cmsx = 0;
          cpl_array_set_int(msx, j, cmsx);
          nrj++;
        }
        if (cnfx < 1. - cstd * lrejt) {
          int cmsx = 0;
          cpl_array_set_int(msx, j, cmsx);
          nrj++;
        }
//        if (cnfx < 1. + cstd * hrejt && cnfx > 1. - cstd * lrejt) {
//          int cmsx = 1;
//          cpl_array_set_int(msx, j, cmsx);
//        }
      }

    }
    //printf("Number of DISCARDED points AFTER: %i \n", nrj);

    nrej = nrj;
    if (!cpl_errorstate_is_equal(prestate)) {
          return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                   "Unable to perform the continuum fitting");
    }
    return CPL_ERROR_NONE;

}


/*----------------------------------------------------------------------------*/
/**
  @brief    Choose order for fitting function for some spectral order
  @param    wax      Wavelength array
  @param    xford    Array with selected order of fitting function
  @return   CPL_ERROR_NONE  iff OK

  This function creates a mask equal to 0 if the wavelength/flux point is rejected.
  The condition for rejection is :
     flx[i]/ctx[i] > 1. + cstd * hrejt
     flx[i]/ctx[i] < 1. - cstd * lrejt
  IF the flux point is not rejected then mask is equal to 1
*/

/*----------------------------------------------------------------------------*/

cpl_error_code espda_cont_ford(cpl_array        *wax,
							   cpl_array      *xford)
{
    cpl_errorstate prestate = cpl_errorstate_get();

    // Defining the number of elements of the arrays
    cpl_size npoint = cpl_array_get_size (wax);

    /* Wavelengths where we have to restrict the order of the fitting polynomial
     * because late stellar type stars have broad and deep features difficult to
     * treat */
    //double waveord1a = 370.0;
    //double waveord1b = 409.0;
    //int ford1=0;
    //double waveord2a = 408.5;
    //double waveord2b = 411.5;
    //int ford2=0;
    //double waveord3a = 429.0;
    //double waveord3b = 435.0;
    //int ford3=1;
    double waveord4a = 484.5;
    double waveord4b = 487.5;
    int ford4=1;
    //double waveord5a = 516.0;
    //double waveord5b = 519.5;
    //int ford5=2;
    double waveord5a = 516.0;
    double waveord5b = 528.5;
    int ford5=1;
    double waveord6a = 654.5;
    double waveord6b = 658.5;
    int ford6=2;
    //double waveord7a = 422.5;
    //double waveord7b = 427.5;
    //int ford7=1;
    //double waveord8a = 436.5;
    //double waveord8b = 442.5;
    //int ford8=1;
    //double waveord9a = 444.5;
    //double waveord9b = 447.5;
    //int ford9=1;

    /*Telluric oxygen features */

	double waveord10a = 765.0;
	//double waveord10a = 755.0;
    double waveord10b = 768.0;
    int ford10=1;

    /* Setting ford for orders with strong features */
    double cwavea = cpl_array_get_double (wax, 0, NULL);
    double cwaveb = cpl_array_get_double (wax, (int)npoint-1, NULL);
    int ford = cpl_array_get_int(xford, 0, NULL);
    int fordera = cpl_array_get_int(xford, 0, NULL);
    int forderb = cpl_array_get_int(xford, 1, NULL);
    int forderc = cpl_array_get_int(xford, 2, NULL);
    int forderd = cpl_array_get_int(xford, 3, NULL);

    //printf("FORD: %i Wmin: %f Wmax: %f \n", ford, cwavea, cwaveb);
    //printf("FORD: %i %i %i %i %i \n", ford, fordera, forderb, forderc, forderd);

    if (cwaveb <= 410.) {
       	ford = fordera;
    } else if (cwaveb <= 450.) {
       	ford = forderb;
    } else if (cwaveb <= 510.) {
      	ford = forderc;
    } else if (cwaveb <= 800.) {
    	ford = forderd;
    }

    //printf("FORD: %i Wmin: %f Wmax: %f \n", ford, cwavea, cwaveb);

    /* Setting ford for orders with strong features */

    if (cwaveb >= waveord4a && cwaveb <= waveord4b) {
      	  ford = ford4;
        } else if (cwavea >= waveord4a && cwavea <= waveord4b) {
      	  ford = ford4;
        } else if (cwavea <= waveord4a && cwaveb >= waveord4b) {
     	  ford = ford4;
        }
    if (cwaveb >= waveord5a && cwaveb <= waveord5b) {
      	  ford = ford5;
        } else if (cwavea >= waveord5a && cwavea <= waveord5b) {
      	  ford = ford5;
        } else if (cwavea <= waveord5a && cwaveb >= waveord5b) {
     	  ford = ford5;
        }
    if (cwaveb >= waveord6a && cwaveb <= waveord6b) {
      	  ford = ford6;
        } else if (cwavea >= waveord6a && cwavea <= waveord6b) {
      	  ford = ford6;
        } else if (cwavea <= waveord6a && cwaveb >= waveord6b) {
     	  ford = ford6;
        }

    if (cwaveb >= waveord10a && cwaveb <= waveord10b) {
          ford = ford10;
        } else if (cwavea >= waveord10a && cwavea <= waveord10b) {
          ford = ford10;
        } else if (cwavea <= waveord10a && cwaveb >= waveord10b) {
          ford = ford10;
        }

    cpl_array_set_int(xford, 4, ford);

    //printf("FORDB: %i Wmin: %f Wmax: %f \n", ford, cwavea, cwaveb);

    if (!cpl_errorstate_is_equal(prestate)) {
          return (int)cpl_error_set_message(cpl_func, cpl_error_get_code(),
                   "Unable to perform the continuum fitting");
    }
    return CPL_ERROR_NONE;

}

/**@}*/

