
/* $Id: espda_compu_eqwidth.c 161960 2014-06-06 14:06:30Z sousasag $
 *
 * This file is part of the ESPDA Pipeline
 * Copyright (C) 2002,2003 European Southern Observatory
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 */

/*
 * $Author: sousasag $
 * $Date: 2014-06-06 15:06:30 +0100 (Sex, 06 Jun 2014) $
 * $Revision: 161960 $
 * $Name: espda_compu_eqwidth $
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#include <string.h>
#endif

/*-----------------------------------------------------------------------------
                                Includes
 -----------------------------------------------------------------------------*/

#include <cpl.h>
#include "cpl_plot.h"

#include "espda_dfs.h"
#include "espda_starpar_lib.h"

/*-----------------------------------------------------------------------------
                             Plugin registration
 -----------------------------------------------------------------------------*/

int cpl_plugin_get_info(cpl_pluginlist * list);

/*-----------------------------------------------------------------------------
                            Private function prototypes
 -----------------------------------------------------------------------------*/

static int espda_compu_starpar_create(cpl_plugin *);
static int espda_compu_starpar_exec(cpl_plugin *);
static int espda_compu_starpar_destroy(cpl_plugin *);
static int espda_compu_starpar(cpl_frameset *, const cpl_parameterlist *);

/*-----------------------------------------------------------------------------
                            Static variables
 -----------------------------------------------------------------------------*/

static char espda_compu_starpar_description[] =
"This recipe computes the effective temperature and the metallicity for solar"
" type stars using quick spectroscopic calibrions."
"\n"
"INPUT FRAMES                        FORMAT  # EXPECTED  TAGS\n" 
"list of lines with EW measurements  LINE    1           LINE_EQWIDTH      \n"
"line ratio calibration fits table   LINE    1           LINE_RATIO_CALIB  \n"
"Fe iron calibration fits table      LINE    1           LINE_FEH_CALIB      "
"\n"
"OUTPUT FRAMES                       FORMAT  # PRODUCED  TAGS                "
"       NOTES\n" 
"effective temperature estimation    TEFF    1           TEFF                "
"       the derived effective temperature in a fits table with errors      \n"
"iron abundance estimation [Fe/H]    FEH     1           FEH                 "
"       the derived iron abundance [Fe/H] in a fits table with errors      \n"
;

/*-----------------------------------------------------------------------------
                                Function code
 -----------------------------------------------------------------------------*/

/**@{*/

/*----------------------------------------------------------------------------*/
/**
  @brief    Build the list of available plugins, for this module. 
  @param    list    the plugin list
  @return   0 if everything is ok, 1 otherwise
  @note     Only this function is exported

  Create the recipe instance and make it available to the application using the 
  interface. 
 */
/*----------------------------------------------------------------------------*/
int cpl_plugin_get_info(cpl_pluginlist *list)
{
    cpl_recipe *recipe = cpl_calloc(1, sizeof *recipe );
    cpl_plugin *plugin = &recipe->interface;

    if (cpl_plugin_init(plugin, CPL_PLUGIN_API, ESPDA_BINARY_VERSION,
                        CPL_PLUGIN_TYPE_RECIPE, "espda_compu_starpar", "Estimate"
                        " the effective temperature and [Fe/H] for solar type star", 
                        espda_compu_starpar_description, "Sergio Sousa",
                        PACKAGE_BUGREPORT, espda_dfs_license(),
                        espda_compu_starpar_create, espda_compu_starpar_exec,
                        espda_compu_starpar_destroy)) {    
        cpl_msg_error(cpl_func, "Plugin initialization failed");
        (void)cpl_error_set_where(cpl_func);                          
        return 1;                                               
    }                                                    

    if (cpl_pluginlist_append(list, plugin)) {                 
        cpl_msg_error(cpl_func, "Error adding plugin to list");
        (void)cpl_error_set_where(cpl_func);                         
        return 1;                                              
    }                                                          
    
    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Setup the recipe options    
  @param    plugin  the plugin
  @return   0 if everything is ok

  Defining the command-line/configuration parameters for the recipe.
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_starpar_create(cpl_plugin *plugin)
{
    cpl_recipe    *recipe;                                               
    cpl_parameter *param;
                                                                     
    /* Do not create the recipe if an error code is already set */     
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                      
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",
                      cpl_func, __LINE__, cpl_error_get_where());      
        return (int)cpl_error_get_code();                              
    }                                                                  
                                                                       
    if (plugin == NULL) {                                              
        cpl_msg_error(cpl_func, "Null plugin");                        
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                 
    }                                                                  
                                                                       
    /* Verify plugin type */                                           
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {       
        cpl_msg_error(cpl_func, "Plugin is not a recipe");             
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);              
    }                                                                  
                                                                       
    /* Get the recipe */                                               
    recipe = (cpl_recipe *)plugin;                                     
                                                                       
    /* Create the parameters list in the cpl_recipe object */          
    recipe->parameters = cpl_parameterlist_new();                      
    if (recipe->parameters == NULL) {                                  
        cpl_msg_error(cpl_func, "Parameter list allocation failed");   
        cpl_ensure_code(0, (int)CPL_ERROR_ILLEGAL_OUTPUT);             
    }                                                                  

    /* Fill the parameters list */
    /* --apply-mask-add */
/*
    param = cpl_parameter_new_value("espda.espda_mask_spec.mask_apply_add",
                                    CPL_TYPE_BOOL, "Flag to reset the mask "
                                    "column before applying the mask",
                                    "espda.espda_mask_spec", TRUE);
    cpl_parameter_set_alias(param, CPL_PARAMETER_MODE_CLI, "apply-mask-add");
    cpl_parameter_disable(param, CPL_PARAMETER_MODE_ENV);
    cpl_parameterlist_append(recipe->parameters, param);
*/

// No parameters for this recipe


    return 0;
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Execute the plugin instance given by the interface
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_starpar_exec(cpl_plugin *plugin)
{
    int             recipe_status;

    cpl_recipe     *recipe = NULL;                                                   
    cpl_errorstate  initial_errorstate = cpl_errorstate_get();              
                                                                           
    /* Return immediately if an error code is already set */               
    if (cpl_error_get_code() != CPL_ERROR_NONE) {                          
        cpl_msg_error(cpl_func, "%s():%d: An error is already set: %s",    
                      cpl_func, __LINE__, cpl_error_get_where());          
        return (int)cpl_error_get_code();                                  
    }                                                                      
                                                                           
    if (plugin == NULL) {                                                  
        cpl_msg_error(cpl_func, "Null plugin");                            
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Verify plugin type */                                               
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {           
        cpl_msg_error(cpl_func, "Plugin is not a recipe");                 
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);                  
    }                                                                      
                                                                           
    /* Get the recipe */                                                   
    recipe = (cpl_recipe *)plugin;                                         
                                                                           
    /* Verify parameter and frame lists */                                 
    if (recipe->parameters == NULL) {                                      
        cpl_msg_error(cpl_func, "Recipe invoked with NULL parameter list");
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
    if (recipe->frames == NULL) {                                          
        cpl_msg_error(cpl_func, "Recipe invoked with NULL frame set");     
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);                     
    }                                                                      
                                                                           
    /* Invoke the recipe */                                                
    recipe_status = espda_compu_starpar(recipe->frames, recipe->parameters);
                                                                           
    /* Ensure DFS-compliance of the products */                            
    if (cpl_dfs_update_product_header(recipe->frames)) {                   
        if (!recipe_status) recipe_status = (int)cpl_error_get_code();                         
    }                                                                      
                                                                           
    if (!cpl_errorstate_is_equal(initial_errorstate)) {                    
        /* Dump the error history since recipe execution start.            
           At this point the recipe cannot recover from the error */       
        cpl_errorstate_dump(initial_errorstate, CPL_FALSE, NULL);          
    }                                                                      
                                                                           
    return recipe_status;                                                  
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Destroy what has been created by the 'create' function
  @param    plugin  the plugin
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_starpar_destroy(cpl_plugin *plugin)
{
    cpl_recipe * recipe;                                          
                                                                  
    if (plugin == NULL) {                                         
        cpl_msg_error(cpl_func, "Null plugin");                   
        cpl_ensure_code(0, (int)CPL_ERROR_NULL_INPUT);            
    }                                                             
                                                                  
    /* Verify plugin type */                                      
    if (cpl_plugin_get_type(plugin) != CPL_PLUGIN_TYPE_RECIPE) {  
        cpl_msg_error(cpl_func, "Plugin is not a recipe");        
        cpl_ensure_code(0, (int)CPL_ERROR_TYPE_MISMATCH);         
    }                                                             
                                                                  
    /* Get the recipe */                                          
    recipe = (cpl_recipe *)plugin;                                
                                                                  
    cpl_parameterlist_delete(recipe->parameters);             
                                                                  
    return 0;                                                    
}

/*----------------------------------------------------------------------------*/
/**
  @brief    Interpret the command line options and execute the data processing
  @param    frameset   the frames list
  @param    parlist    the parameters list
  @return   0 if everything is ok
 */
/*----------------------------------------------------------------------------*/
static int espda_compu_starpar(cpl_frameset   *frameset,
			                   const cpl_parameterlist *paramlist)
{
//    cpl_errorstate    prestate = cpl_errorstate_get();
    cpl_table        *line_ew_table = NULL;
    cpl_table        *lineratio_calib_table = NULL;
    cpl_table        *linefeh_calib_table = NULL;

    cpl_table        *teff_table = NULL;
    cpl_table        *feh_table = NULL;

    cpl_frame        *line_ew_frame = NULL;                                            
    cpl_frame        *lineratio_calib_frame = NULL;                                            
    cpl_frame        *linefeh_calib_frame = NULL;                                            

// iteration cariables

    cpl_msg_info(cpl_func, "*** Entering Compu Star Par ***");

    cpl_msg_info(cpl_func, "*** Checking Line EW frame  ***");


/*
 * DFS groups are set for all input frames and the data are loaded into a 
 * merged table (for 1D spectra) and/or an order table (for 2D spectra). 
 * Functions espda_dfs_groups() and espda_open_frames() are called.
 * Property list for QC parameters is created.
 */

    cpl_msg_info(cpl_func, "*** Set groups of FITS frames ***");
    cpl_ensure_code(espda_dfs_groups(frameset) == CPL_ERROR_NONE,
                    cpl_error_get_code());



// Getting the 1D spec fits table frame
    cpl_frame        *spec_frame = NULL;
    spec_frame = cpl_frameset_find(frameset, RSPEC_PRE);
// Checking if the MERGESPEC_TAB fits frame is defined
    if (spec_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func,
               CPL_ERROR_DATA_NOT_FOUND, "The S1D RSPEC_PRE FITS"
               "TABLE is not defined: Need a frame with RSPEC_PRE");
    }

    // Loading into a cpl table
    cpl_table        *spec_table = NULL;
    spec_table = cpl_table_load( cpl_frame_get_filename(spec_frame), 1, 0);


// Getting the LINE_EQWIDTH table frame
    line_ew_frame = cpl_frameset_find(frameset, LINE_EQWIDTH);
// Checking if the MERGESPEC_TAB fits frame is defined
    if (line_ew_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_DATA_NOT_FOUND, "The LINE_EQWIDTH" 
               "TABLE is not defined: Need a frame with LINE_EQWIDTH with the EWs");
    }

// Getting the keyword in the frame to check if is really a MERGESPEC
    cpl_propertylist* proplist = cpl_propertylist_load_regexp(
	                     cpl_frame_get_filename(line_ew_frame), 0, CPL_DFS_PRO_CATG, 0);
// Check if Keyword is present here:
    if (proplist == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE_EQWIDTH" 
               "TABLE is not defined: Need a frame with LINE_EQWIDTH with the EWs");
    }
        
// Check if Keyword is compatible
    if (strcmp(espda_dfs_catg(proplist),LINE_EQWIDTH)!=0){
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE FITS" 
               "File is not a LINE_EQWIDTH Frame - Keyword not compatible");
    }
    cpl_propertylist_delete(proplist);
    cpl_msg_info(cpl_func, "*** LINE_EQWIDTH frame OK***");

// -------------------------------------------------------------------

    cpl_msg_info(cpl_func, "*** Checking LINE Ratio Calib TABLE frame ***");

// Getting the 1D spec fits table frame
    lineratio_calib_frame = cpl_frameset_find(frameset, LINE_RATIO_CALIB);
// Checking if the LINE LIST fits frame is defined
    if (lineratio_calib_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_DATA_NOT_FOUND, "The LINE Ratio Calib FITS " 
               "TABLE is not defined: Need a frame with LINE_RATIO_CALIB");
    }

// Getting the keyword in the frame to check if is really a LINE
    proplist = cpl_propertylist_load_regexp(
	                     cpl_frame_get_filename(lineratio_calib_frame), 0, CPL_DFS_PRO_CATG, 0);

// Check if Keyword is present:
    if (proplist == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE Ratio Calib FITS " 
               "File is not a LINE_RATIO_CALIB Frame - Keyword missing");
    }
        
// Check if Keyword is compatible
    if (strcmp(espda_dfs_catg(proplist),LINE_RATIO_CALIB)!=0){
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE Ratio Calib FITS" 
               "File is not a LINE_RATIO_CALIB Frame - Keyword not compatible");
    }
    cpl_propertylist_delete(proplist);
    cpl_msg_info(cpl_func, "*** LINE RATIO CALIB frame OK***");

// -------------------------------------------------------------------


    cpl_msg_info(cpl_func, "*** Checking LINE FEH Calib TABLE frame ***");

// Getting the 1D spec fits table frame
    linefeh_calib_frame = cpl_frameset_find(frameset, LINE_FEH_CALIB);
// Checking if the LINE LIST fits frame is defined
    if (linefeh_calib_frame == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_DATA_NOT_FOUND, "The LINE FEH Calib FITS" 
               "TABLE is not defined: Need a frame with LINE_FEH_CALIB");
    }

// Getting the keyword in the frame to check if is really a LINE
    proplist = cpl_propertylist_load_regexp(
	                     cpl_frame_get_filename(linefeh_calib_frame), 0, CPL_DFS_PRO_CATG, 0);

// Check if Keyword is present:
    if (proplist == NULL) {
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE FEH Calib FITS" 
               "File is not a LINE_FEH_CALIB Frame - Keyword missing");
    }
        
// Check if Keyword is compatible
    if (strcmp(espda_dfs_catg(proplist),LINE_FEH_CALIB)!=0){
        return (int)cpl_error_set_message(cpl_func, 
               CPL_ERROR_BAD_FILE_FORMAT, "The LINE FEH Calib FITS" 
               "File is not a LINE_FEH_CALIB Frame - Keyword not compatible");
    }
    cpl_propertylist_delete(proplist);
    cpl_msg_info(cpl_func, "*** LINE FEH CALIB frame OK***");

// -------------------------------------------------------------------



    // Loading cpl table
    line_ew_table = cpl_table_load( cpl_frame_get_filename(line_ew_frame), 1, 0);
//    printf("Filename line EW Fits Table: %s\n", cpl_frame_get_filename(line_ew_frame));

    // Loading the calibrations
    lineratio_calib_table = cpl_table_load(cpl_frame_get_filename(lineratio_calib_frame), 1, 0);
//    printf("Filename ratio calib Fits Table: %s\n", cpl_frame_get_filename(lineratio_calib_frame));

    linefeh_calib_table = cpl_table_load(cpl_frame_get_filename(linefeh_calib_frame), 1, 0);
    //printf("Filename FEH calib Fits Table: %s\n", cpl_frame_get_filename(linefeh_calib_frame));
  

//    cpl_table_dump(lineratio_calib_table,0,10,NULL);	

    cpl_msg_info(cpl_func, "*** Computing the Teff ***");
    cpl_ensure_code(espda_calib_teff(lineratio_calib_table, 
                                   line_ew_table, 
                                   &teff_table)
                        == CPL_ERROR_NONE, cpl_error_get_code());


/*    cpl_ensure_code(esp_calib_feh(linefeh_calib_table, 
                                    line_ew_table, 
                                    Teffout, erTeffout1, erTeffout2, erTeffout3, 
                                    &feh, &erfeh,&nfeh)
                        == CPL_ERROR_NONE, cpl_error_get_code());
*/                        
    cpl_ensure_code(espda_calib_feh(linefeh_calib_table, 
                                    line_ew_table,
                                    teff_table,
                                    &feh_table)
                        == CPL_ERROR_NONE, cpl_error_get_code());



    // Create property lists
    proplist = cpl_propertylist_new();
    cpl_propertylist_append_double(proplist, DAS_TEFF, cpl_table_get_double(teff_table, "TEFF",0,NULL));
    cpl_propertylist_append_double(proplist, DAS_STDTEFF, cpl_table_get_double(teff_table, "STD_TEFF",0,NULL));
    cpl_propertylist_append_int(proplist, DAS_NCALTEFF, cpl_table_get_int(teff_table, "N_CALIB",0,NULL));
    cpl_propertylist_append_int(proplist, DAS_NCALINDTEFF, cpl_table_get_int(teff_table, "N_CALIB_IND",0,NULL));
    cpl_propertylist_append_double(proplist, DAS_FEH, cpl_table_get_double(feh_table, "FEH",0,NULL));
    cpl_propertylist_append_double(proplist, DAS_STDFEH, cpl_table_get_double(feh_table, "STD_FEH",0,NULL));
    cpl_propertylist_append_int(proplist, DAS_NCALFEH, cpl_table_get_int(feh_table, "N_CALIB",0,NULL));

    cpl_propertylist_append_string(proplist, CPL_DFS_PRO_CATG, "RSPEC");
    cpl_ensure_code(
        cpl_dfs_save_table(frameset, NULL, paramlist, frameset,
                           NULL, spec_table, NULL,
                           "espda_compu_starpar", proplist, NULL, PACKAGE "/"
                           PACKAGE_VERSION, "RSPEC.fits")
        == CPL_ERROR_NONE, cpl_error_get_code()
    );

/*
// Saving the product as a fits table
    proplist = cpl_propertylist_new();
    cpl_propertylist_append_string(proplist, CPL_DFS_PRO_CATG, "TEFF");
    cpl_ensure_code(
        cpl_dfs_save_table(frameset, NULL, paramlist, frameset, 
                           NULL, teff_table, NULL, 
                           "espda_compu_starpar", proplist, NULL, PACKAGE "/"
                           PACKAGE_VERSION, "TEFF_EW.fits")          
        == CPL_ERROR_NONE, cpl_error_get_code()
    );
    cpl_propertylist_delete(proplist);


// Saving the product as a fits table
    proplist = cpl_propertylist_new();
    cpl_propertylist_append_string(proplist, CPL_DFS_PRO_CATG, "FEH");
    cpl_ensure_code(
        cpl_dfs_save_table(frameset, NULL, paramlist, frameset, 
                           NULL, feh_table, NULL, 
                           "espda_compu_starpar", proplist, NULL, PACKAGE "/"
                           PACKAGE_VERSION, "FEH_EW.fits")          
        == CPL_ERROR_NONE, cpl_error_get_code()
    );
    cpl_propertylist_delete(proplist);

*/
// Final Memory Free
//    cpl_propertylist_delete (proplist);
    cpl_table_delete(line_ew_table);
    cpl_table_delete(lineratio_calib_table);
    cpl_table_delete(linefeh_calib_table);
    cpl_table_delete(teff_table);
    cpl_table_delete(feh_table);
    cpl_table_delete(spec_table);



    cpl_msg_info(cpl_func, "*** Success!! ***");
    return (int)cpl_error_get_code();

}

/**@}*/
